#!/usr/bin/env python
'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''


import argparse
from cvss import CVSS2, CVSS3

import sys
import importlib

from text_office.utils import log

parser = argparse.ArgumentParser()
parser.add_argument('module', help='boilerplate generator module to use (e.g., vapt)', type=str)
args = parser.parse_args(sys.argv[1:2])

try:
    importstr = f'text_office.boilers.{args.module}'
    boiler = importlib.import_module(importstr)
    boiler.generate(sys.argv[2:])
except ImportError:
    log.exception(f'boilerplate module \'{args.module}\' not found.')
except Exception as e:
    log.exception('exception occured')



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

# special formatting tokens to ease report writing to md

# https://github.com/miyuchina/mistletoe/blob/master/mistletoe/html_renderer.py

import re
from ..utils import parse_kv_pairs
from mistletoe.span_token import SpanToken, RawText
from mistletoe.block_token import BlockToken, tokenize

_attr_matcher = r'([^\<\>\n]+)'

def _build_regex_ftag_uni_pattern(tag):
    return f'<{tag}\\s*{_attr_matcher}?>'

def _build_regex_ftag_pattern(tag):
    return f'<{tag}\\s*{_attr_matcher}?>(.*?)</{tag}>'

def _build_regex_ftag_noattr_pattern(tag):
    return f'<{tag}>(.*?)</{tag}>'

def _build_regex_ftag_start_pattern(tag):
    return f'<{tag}\\s*{_attr_matcher}?[>\n]'
    #return f'<{tag}\\s*(\\S.*?)?>\n'

class KeyValueMixin:
    def parse_format(self):
        self.format = {}
        if self.format_value_raw == '>':
            # edge case 1, KeyValueFormatTag without attributes
            pass
        elif self.format_value_raw is None:
            # edge case 2, ditto
            pass
        else:
            self.format = parse_kv_pairs(self.format_value_raw, item_sep=' \n\t')


class NoAttrFormatTag(SpanToken):
    parse_group = 1


class FormatTag(SpanToken):
    parse_group = 2

    def __init__(self, match):
        self.format_value = None
        self.format_value_raw = match.group(1)
        if self.format_value_raw is not None:
            self.format_value = self.format_value_raw.casefold()
        # do not define children ourselves, allow for nested formats
        #self.children = (RawText(match.group(2) if match.group(2) is not None else match.group(4)),)


class NoBodyFormatTag(FormatTag):
    parse_group = 0
    parse_inner = False


class KeyValueFormatTag(KeyValueMixin, FormatTag):
    def __init__(self, match):
        super().__init__(match)
        self.parse_format()

class KeyValueNoBodyFormatTag(KeyValueMixin, NoBodyFormatTag):
    def __init__(self, match):
        super().__init__(match)
        self.parse_format()

class BoldTag(NoAttrFormatTag):
    pattern = re.compile(_build_regex_ftag_noattr_pattern('b'), re.DOTALL)

class ItalicTag(NoAttrFormatTag):
    pattern = re.compile(_build_regex_ftag_noattr_pattern('i'), re.DOTALL)

class UnderlineTag(NoAttrFormatTag):
    pattern = re.compile(_build_regex_ftag_noattr_pattern('u'), re.DOTALL)

class StrongTag(NoAttrFormatTag):
    pattern = re.compile(_build_regex_ftag_noattr_pattern('strong'), re.DOTALL)

class EmphasisTag(NoAttrFormatTag):
    pattern = re.compile(_build_regex_ftag_noattr_pattern('emphasis'), re.DOTALL)

class StrikethroughTag(NoAttrFormatTag):
    pattern = re.compile(_build_regex_ftag_noattr_pattern('strike'), re.DOTALL)

class FontTag(KeyValueFormatTag):
    pattern = re.compile(_build_regex_ftag_pattern('font'), re.DOTALL)

class ImageTag(KeyValueFormatTag):
    pattern = re.compile(_build_regex_ftag_pattern('img'), re.DOTALL)

class CellTag(KeyValueFormatTag):
    pattern = re.compile(_build_regex_ftag_pattern('cell'), re.DOTALL)

class HorizontalRuleTag(NoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('hr'))

class LineBreakTag(NoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('br'))

class SectionBreakTag(NoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('secbr'))

class InsertTabTag(NoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('tab'))

class PageBreakTag(NoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('pgbr'))

class TOCTag(NoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('toc'))

class LOTTag(NoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('lot'))

class LOFTag(NoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('lof'))

class MergeTag(KeyValueNoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('merge'))

class InsertPageNumTag(KeyValueNoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('pgnum'))

class SectionControlTag(KeyValueNoBodyFormatTag):
    pattern = re.compile(_build_regex_ftag_uni_pattern('section'))

class NoNewLineException(Exception):

    def __init__(self, tagname):
        super().__init__(f'invalid block tag "{tagname}": can\'t find end tag, did you include a newline after the start tag/before the end tag (at least 2 lines required)?')

class ParseException(Exception):
    def __init__(self, tagname, start, lines):
        super().__init__(f'failed to parse {tagname} tag: "{start}" with lines {lines}')

class FormatBlockTag(BlockToken):

    _end_cond = None

    def __init__(self, lines):
        try:
            start_token = re.search(_build_regex_ftag_start_pattern(self.tag), lines[0], re.DOTALL)
            self.format_value_raw = start_token.group(1)
            if self.format_value_raw is not None:
                self.format_value = self.format_value_raw.casefold()
            lines[0] = lines[0][start_token.span()[1]:]
            if len(lines[-1].strip()) < 1:
                raise NoNewLineException(self.tag)
            end_token =re.search(f'</{self.tag}>', lines[-1])
            lines[-1] = lines[-1][:end_token.span()[0]]
            super().__init__(lines, tokenize)
        except Exception as e:
            raise ParseException(self.tag, start_token.group(0), ''.join(lines))

    @classmethod
    def start(cls, line):
        stripped = line.lstrip()
        if len(line) - len(stripped) >= 4:
            return False

        # rule 1: <pre>, <script> or <style> tags, allow newlines in block
        match_obj = re.match(_build_regex_ftag_start_pattern(cls.tag), stripped, re.DOTALL)
        cls._end_cond = f'</{cls.tag}>'
        if match_obj is not None:
            return 1
        return False

    @classmethod
    def read(cls, lines):
        # note: stop condition can trigger on the starting line
        line_buffer = []
        for line in lines:
            line_buffer.append(line)
            if cls._end_cond is not None:
                if cls._end_cond in line.casefold():
                    break
            elif line.strip() == '':
                line_buffer.pop()
                break
        return line_buffer

class FormatKeyValueBlockTag(KeyValueMixin, FormatBlockTag):
    def __init__(self, lines):
        super().__init__(lines)
        self.parse_format()

# comment block
class CommentBlockTag(FormatBlockTag):
    tag = 'comment'

# stuff that require multiple lines with breaks in between (e.g., align multiple paragraphs, styling a multi-line table)
class AlignBlockTag(FormatBlockTag):
    tag = 'align'

class TableBlockTag(FormatKeyValueBlockTag):
    tag = 'table'

class BorderBlockTag(FormatKeyValueBlockTag):
    tag = 'border'

class ParagraphBlockTag(FormatKeyValueBlockTag):
    tag = 'para'

class HeaderBlockTag(FormatKeyValueBlockTag):
    tag = 'header'

class FooterBlockTag(FormatKeyValueBlockTag):
    tag = 'footer'

class FontBlockTag(FormatKeyValueBlockTag):
    tag = 'param'



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import argparse
import os
import sys
from string import Template
from .. import utils

def generate(inargs):
    parser = argparse.ArgumentParser(description="static boilerplate module")
    parser.add_argument('template', help='template (relative to boiler_templates dir. in project root)', type=str)
    parser.add_argument('-d', '--data', help='key-value pair of substitutions options (e.g., foo=bar, count=3)', type=str)
    parser.add_argument('-o', '--output', help='output md filename', type=str, default='output.md')
    parser.add_argument('-stdout', '--stdout', help='output to stdout', action="store_true")
    parser.add_argument('--overwrite', help='overwrite output instead of append', action='store_true')
    args = parser.parse_args(inargs)

    kvd = {}
    if args.data is not None:
        kvd = utils.parse_kv_pairs(args.data)

    tplfn = utils.ensure_template_file(args.template, auto_add_suffix='.md')

    buf = []
    with open(tplfn, 'r') as tpl:
        buf = tpl.read()

    t = Template(buf)
    buf = t.substitute(**kvd)

    if args.stdout:
        sys.stdout.write(buf)
    else:
        utils.log.info(f'creating boilerplate markdown from "{tplfn}" with values: {kvd}')
        mode = 'w' if args.overwrite else 'a+'
        with open(args.output, mode) as out:
            out.write(buf)



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import os
from .path_helper import (
        boiler_template_path,
        boiler_template_path_pip,
        )

from .log_helper import log


class InvalidAttr(KeyError):
    def __init__(self, attr, valids):
        super().__init__(f'invalid attr: {attr} {valids}')

class MissingAttr(KeyError):
    def __init__(self, attr, required):
        super().__init__(f'missing attr: {attr} {required}')

class InvalidValue(ValueError):
    def __init__(self, attr, value, valids):
        super().__init__(f'invalid value on attr "{attr}": {value} {valids}')

class InvalidType(TypeError):
    def __init__(self, used, desired):
        super().__init__(f'invalid type used "{used}", requires "{desired}"')

def ensure_and_set(vkey, vtype, v, obj, objkey=None, parser=None):
    if objkey is None:
        objkey = vkey

    if isinstance(v, dict):
        v = v.get(vkey)

    if callable(parser):
        v = parser(v)

    if isinstance(vtype, list):
        # allow OR types
        for vtype_or in vtype:
            if isinstance(v, vtype_or):
                # if any meets, sets and exit
                setattr(obj, objkey, v)
                return
    else:
        # singular type
        if isinstance(v, vtype):
            # meets, sets and exists
            setattr(obj, objkey, v)
            return

    if v is not None:
        # v is specified, but invalid type, dev error
        raise InvalidType(type(v), vtype)

def ensure_all_attr(required, opt_dict):
    for k in required.keys():
        if k not in opt_dict:
            raise MissingAttr(k, required)
        parser, valids = required[k]
        v = opt_dict.get(k)
        try:
            opt_dict[k] = parser(v)
        except Exception as e:
            raise InvalidValue(k, v, valids)

def ensure_valid_attr(valid_list, opt_dict):
    for k in opt_dict:
        if k not in valid_list:
            raise InvalidAttr(k, valid_list)


def ensure_valid_value(attr, vmap, v):
    if v not in vmap:
        raise InvalidValue(attr, v, list(vmap.keys()))


def ensure_template_file(t, auto_add_suffix=None):

    tplfn = t
    if isinstance(auto_add_suffix, str):
        if not tplfn.endswith(auto_add_suffix):
            tplfn += auto_add_suffix

    btplfn = tplfn

    if not os.path.isfile(tplfn):
        log.warn(f"cannot find boiler template '{tplfn}' in current working directory, defaulting to built-ins")
        tplfn = os.path.join(boiler_template_path, btplfn)

    if not os.path.isfile(tplfn):
        tplfn = os.path.join(boiler_template_path_pip, btplfn)

    if not os.path.isfile(tplfn):
        raise FileExistsError(f"cannot find boiler template '{btplfn}' in built-in templates dir")

    return tplfn



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import logging
import sys

#log_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
log_formatter = logging.Formatter('%(message)s')
log_handler = logging.StreamHandler(sys.stdout)
log_handler.setLevel(logging.DEBUG)
log_handler.setFormatter(log_formatter)

log = logging.getLogger()
log.setLevel(logging.DEBUG)
log.addHandler(log_handler)



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from cvss import CVSS2, CVSS3

def get_riskcolor(severity):
    _map = {
            'critical':'#c00000',
            'high':'#ff0000',
            'medium':'#ffc000',
            'low':'#92d050',
            'info':'#00b0f0'
    }
    return _map[severity.casefold()]

def get_severity(score):
    if score >= 9.0:
        return 'Critical'
    elif score >= 7.0:
        return 'High'
    elif score >= 4.0:
        return 'Medium'
    elif score >= 0.1:
        return 'Low'
    else:
        return 'Info'

def get_ssvc(cvss):

    if cvss is None:
        # info
        severity = 'Info'
        score = 0.0
        vector = '-'
    else:
        if not cvss.startswith('CVSS:3.1/'):
            cvss = 'CVSS:3.1/'+cvss
        _cvss_o = CVSS3(cvss)

        #severity = _cvss_o.severities()[0]
        score = min(_cvss_o.scores())
        severity = get_severity(score)
        vector = _cvss_o.clean_vector()

    if score < 0.1:
        score_print = '-'
    else:
        score_print = str(score)

    color = get_riskcolor(severity)

    return severity, score, score_print, vector, color



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import os
import sys

_path_ = os.path.realpath(__file__)
_base_, _file_ = os.path.split(_path_)
_lib_ = _base_[:-_base_[::-1].find(os.path.sep)-1]
prj_root_path = _lib_[:-_lib_[::-1].find(os.path.sep)]
boiler_template_path = os.path.join(prj_root_path, 'boiler_templates')

_mainmod_path_ = os.path.abspath(sys.modules['__main__'].__file__)
_mainmod_base_, _mainmod_file_ = os.path.split(_mainmod_path_)
_local_path_ = _mainmod_base_[:-_mainmod_base_[::-1].find(os.path.sep)-1]
boiler_template_path_pip = os.path.join(_local_path_, 'boiler_templates')



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import webcolors
import re
from shlex import shlex
from docx.shared import Pt, Inches, Cm, Mm, RGBColor, Length
from docx.enum.section import WD_SECTION, WD_ORIENT
from docx.enum.text import WD_ALIGN_PARAGRAPH, WD_TAB_ALIGNMENT
from docx.enum.table import WD_TABLE_ALIGNMENT
from .errx_helper import ensure_valid_value

def parse_color(color):
    if color is None:
        return None

    if color.startswith('#') and len(color) > 6:
        try:
            return RGBColor.from_string(color[1:])
        except Exception as e:
            pass

    try:
        return RGBColor.from_string(webcolors.name_to_hex(color)[1:])
    except Exception as e:
        pass

    try:
        return RGBColor.from_string(color)
    except Exception as e:
        pass

    raise ValueError(f'unable to parse color: {color}')


def parse_sizespec(sizespec):
    if sizespec is None:
        return None

    try:
        match = re.search('([0-9.]+)([a-z]*)', sizespec)
        rsval = float(match.group(1))
        rstyp = match.group(2)

        if rstyp == 'mm':
            return Mm(rsval)
        elif rstyp == 'pt':
            return Pt(rsval)
        elif rstyp == 'in':
            return Inches(rsval)
        elif rstyp == 'cm':
            return Cm(rsval)
        else:
            return rsval
    except Exception as e:
        raise ValueError(f'unable to parse size spec: {sizespec}')


def parse_sec_orientation(orientation):
    vmap = {
        'landscape': WD_ORIENT.LANDSCAPE,
        'portrait': WD_ORIENT.PORTRAIT,
        None: None,
    }
    ensure_valid_value('orientation', vmap, orientation)
    return vmap[orientation]


def parse_para_align(align):
    vmap = {
        'center': WD_ALIGN_PARAGRAPH.CENTER,
        'left': WD_ALIGN_PARAGRAPH.LEFT,
        'right': WD_ALIGN_PARAGRAPH.RIGHT,
        'justify': WD_ALIGN_PARAGRAPH.JUSTIFY,
        None: None
    }
    ensure_valid_value('align', vmap, align)
    return vmap[align]


def parse_table_align(align):
    vmap = {
            'center': WD_TABLE_ALIGNMENT.CENTER,
            'left': WD_TABLE_ALIGNMENT.LEFT,
            'right': WD_TABLE_ALIGNMENT.RIGHT,
            None: None
            }
    ensure_valid_value('align', vmap, align)
    return vmap[align]


def parse_kv_pairs(text, item_sep=", ", value_sep="="):
    """Parse key-value pairs from a shell-like text"""
    # initialize a lexer, in POSIX mode (to properly handle escaping)
    lexer = shlex(text, posix=True)
    # set ',' as whitespace for the lexer
    # (the lexer will use this character to separate words)
    lexer.whitespace = item_sep
    # include '=' as a word character
    # (this is done so that the lexer returns a list of key-value pairs)
    # (if your option key or value contains any unquoted special character, you will need to add it here)
    lexer.wordchars += value_sep
    lexer.wordchars += '!@#$%^&*()[];:.<>?/+-' # allow dot, slahes, round and square brackets
    # then we separate option keys and values to build the resulting dictionary
    # (maxsplit is required to make sure that '=' in value will not be a problem)

    try:
        od = dict(word.split(value_sep, maxsplit=1) for word in lexer)
        return od
    except Exception as e:
        raise ValueError(f'attribute error: {text}: {e}')


def parse_bool(val):
    if isinstance(val, str):
        val = val.casefold()
    if val in [1, True, '1', 'yes', 'true']:
        return True
    return False



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import os
import sys
import traceback
from .errx_helper import (
        ensure_valid_value,
        ensure_template_file,
        )

from .log_helper import log

from .path_helper import (
        boiler_template_path_pip,
        boiler_template_path,
        )

from .parsers import (
        parse_bool,
        parse_kv_pairs,
        parse_sizespec as parse_docx_sizespec,
        )

def set_attr_recursively(token, instance_type, attr, value):
    if isinstance(token, instance_type):
        setattr(token, attr, value)

    if hasattr(token, 'children'):
        if isinstance(token.children, list):
            for c in token.children:
                set_attr_recursively(c, instance_type, attr, value)

default_encoding = 'utf-8'



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import re
import lxml
from .errx_helper import (
        ensure_valid_attr,
        ensure_valid_value,
        ensure_and_set,
        ensure_all_attr,
        )

from docx import Document
from docx.table import Table
from docx.text.paragraph import Paragraph
from docxcompose.composer import Composer
from docx.oxml.shared import OxmlElement
from docx.oxml.ns import qn, nsdecls
from docx.oxml import parse_xml, CT_P
from docx.opc.constants import RELATIONSHIP_TYPE
from docx.oxml.numbering import CT_Num, CT_Numbering, CT_NumPr
from docx.enum.section import WD_SECTION, WD_ORIENT
from docx.enum.text import WD_ALIGN_PARAGRAPH, WD_TAB_ALIGNMENT
from docx.enum.base import XmlMappedEnumMember, EnumValue
from docx.enum.dml import MSO_THEME_COLOR_INDEX
from docx.shared import (
        Length, Pt, Inches, Cm, Mm, Twips,
        RGBColor,
        )
from docx.styles.styles import Styles
from docx.oxml.xmlchemy import serialize_for_reading
from .parsers import (
        parse_sizespec,
        parse_color,
        parse_sec_orientation,
        parse_table_align,
        parse_para_align,
        parse_bool,
        )

def insert_section(doc, stype):
    vmap = {
        'oddpage': WD_SECTION.ODD_PAGE,
        'evenpage': WD_SECTION.EVEN_PAGE,
        'newpage': WD_SECTION.NEW_PAGE,
        'newcol': WD_SECTION.NEW_COLUMN,
        'cont': WD_SECTION.CONTINUOUS,
        None: WD_SECTION.NEW_PAGE,
    }
    ensure_valid_value('secbr', vmap, stype)
    return doc.add_section(vmap[stype])

def merge_table_cells(table, **kwargs):
    _optmap = {
            'from_row': (int, 'digits'),
            'from_col': (int, 'digits'),
            'to_row': (int, 'digits'),
            'to_col': (int, 'digits'),
            }
    ensure_all_attr(_optmap, kwargs)
    fr = kwargs.get('from_row')
    fc = kwargs.get('from_col')
    tr = kwargs.get('to_row')
    tc = kwargs.get('to_col')

    if tr < fr:
        fr, tr = tr, fr

    if tc < fc:
        fc, tc = tc, fc

    merge_tar = table.cell(fr, fc)
    merge_tar.merge(table.cell(tr, tc))
    return table

def format_section(section, **kwargs):
    _optmap = {
            'left_margin':  ([Length, float], kwargs, section, None, parse_sizespec),
            'right_margin':  ([Length, float], kwargs, section, None, parse_sizespec),
            'top_margin':  ([Length, float], kwargs, section, None, parse_sizespec),
            'bottom_margin':  ([Length, float], kwargs, section, None, parse_sizespec),
            'page_width':  ([Length, float], kwargs, section, None, parse_sizespec),
            'page_height':  ([Length, float], kwargs, section, None, parse_sizespec),
            'orientation': (EnumValue, kwargs, section, None, parse_sec_orientation),
            'header_distance': ([Length, float], kwargs, section, None, parse_sizespec),
            'footer_distance': ([Length, float], kwargs, section, None, parse_sizespec),
            'header_linked': (bool, kwargs, section.header, 'is_linked_to_previous', parse_bool),
            'footer_linked': (bool, kwargs, section.footer, 'is_linked_to_previous', parse_bool)
            }
    ensure_valid_attr(_optmap.keys(), kwargs.keys())

    for k, v in _optmap.items():
        if v is None:
            # special cases
            continue
        ensure_and_set(k, *v)

    # force orientation change by swapping page_width and page_height
    page_orient = parse_sec_orientation(kwargs.get('orientation'))
    if isinstance(page_orient, EnumValue):
        # force orientation change
        pw = section.page_width
        section.page_width = section.page_height
        section.page_height = pw

    return section

def format_run(run, **kwargs):
    _optmap = {
            'bold':         (bool, kwargs, run.font, None, parse_bool),
            'italic':       (bool, kwargs, run.font, None, parse_bool),
            'underline':    (bool, kwargs, run.font, None, parse_bool),
            'strike':       (bool, kwargs, run.font, None, parse_bool),
            'name':         (str, kwargs, run.font),
            'style':        (str, kwargs, run),
            'size':         ([Length, float], kwargs, run.font, None, parse_sizespec),
            'color':        (RGBColor, kwargs, run.font.color, 'rgb', parse_color),
            }
    ensure_valid_attr(_optmap.keys(), kwargs.keys())

    for k, v in _optmap.items():
        if v is None:
            # special cases
            continue
        ensure_and_set(k, *v)
    return run

def format_paragraph(para, **kwargs):
    _optmap = {
            'style': (str, kwargs, para.style),
            'align': (EnumValue, kwargs, para.paragraph_format, 'alignment', parse_para_align),
            'spacing': ([Length, float], kwargs, para.paragraph_format, 'line_spacing', parse_sizespec),
            'before': ([Length, float], kwargs, para.paragraph_format, 'space_before', parse_sizespec),
            'after': ([Length, float], kwargs, para.paragraph_format, 'space_after', parse_sizespec),
            'left_indent': ([Length, float], kwargs, para.paragraph_format, None, parse_sizespec),
            'right_indent': ([Length, float], kwargs, para.paragraph_format, None, parse_sizespec),
            'first_line_indent': ([Length, float], kwargs, para.paragraph_format, None, parse_sizespec),
            'tabstops': None,
            }
    ensure_valid_attr(_optmap.keys(), kwargs.keys())

    for k, v in _optmap.items():
        if v is None:
            # special cases
            continue
        ensure_and_set(k, *v)

    tabstops = kwargs.get('tabstops')
    if isinstance(tabstops, list):
        for ats in tabstops:
            para.paragraph_format.tab_stops.add_tab_stop(*ats)

    return para

def format_figure(figobj, **kwargs):
    _optmap = ('width', 'height', 'border_width', 'border_color')
    ensure_valid_attr(_optmap, kwargs.keys())
    set_figure_dims(figobj, kwargs.get('width'), kwargs.get('height'))

    bw = kwargs.get('border_width')
    bc = kwargs.get('border_color')
    if bw is not None:
        if bc is not None:
            set_figure_border(figobj, bw, bc)
        else:
            set_figure_border(figobj, bw)

def format_table(table, **kwargs):
    _optmap = {
            'style': (str, kwargs, table),
            'align': (EnumValue, kwargs, table, 'alignment', parse_table_align),
            'autofit': (bool, kwargs, table, None, parse_bool),
            'column_widths': None,
            'left_indent': None,
            }
    ensure_valid_attr(_optmap.keys(), kwargs.keys())

    for k, v in _optmap.items():
        if v is None:
            # special cases
            continue
        ensure_and_set(k, *v)

    colwidths = kwargs.get('column_widths')
    if isinstance(colwidths, str):
        cwidths=[]
        for cr in colwidths.split(','):
            cwidths.append(parse_sizespec(cr.strip()))

        table.autofit = False
        for r in table.rows:
            for idx, cell in enumerate(r.cells):
                if idx < len(cwidths):
                    cell.width = cwidths[idx]

    left_indent = parse_sizespec(kwargs.get('left_indent'))
    if left_indent is not None:
        indent_table(table, left_indent)

    return table

def insert_hrule(para, linestyle='single'):
    p = para._p  # p is the <w:p> XML element
    pPr = p.get_or_add_pPr()
    pBdr = OxmlElement('w:pBdr')
    pPr.insert_element_before(pBdr,
            'w:shd', 'w:tabs', 'w:suppressAutoHyphens', 'w:kinsoku', 'w:wordWrap',
            'w:overflowPunct', 'w:topLinePunct', 'w:autoSpaceDE', 'w:autoSpaceDN',
            'w:bidi', 'w:adjustRightInd', 'w:snapToGrid', 'w:spacing', 'w:ind',
            'w:contextualSpacing', 'w:mirrorIndents', 'w:suppressOverlap', 'w:jc',
            'w:textDirection', 'w:textAlignment', 'w:textboxTightWrap',
            'w:outlineLvl', 'w:divId', 'w:cnfStyle', 'w:rPr', 'w:sectPr',
            'w:pPrChange'
            )
    bottom = OxmlElement('w:bottom')
    vmap = {
            'dashsmall': 'dashSmallGap',
            'single': 'single',
            None: 'single',
            }
    ensure_valid_value('hr', vmap, linestyle)
    bottom.set(qn('w:val'), vmap[linestyle])
    bottom.set(qn('w:sz'), '6')
    bottom.set(qn('w:space'), '1')
    bottom.set(qn('w:color'), 'auto')
    pBdr.append(bottom)
    return pBdr

def format_tabstops(tabstop, cs):
    if tabstop is None:
        return None

    lmar = Inches(cs.left_margin.inches)
    cmar = Inches(cs.page_width.inches/2 - cs.left_margin.inches)
    rmar = Inches(cs.page_width.inches - (cs.left_margin.inches + cs.right_margin.inches))
    vmap = {
        'left': (lmar, WD_TAB_ALIGNMENT.LEFT),
        'center': (cmar, WD_TAB_ALIGNMENT.CENTER),
        'right': (rmar, WD_TAB_ALIGNMENT.RIGHT),
    }
    out = []
    for t in tabstop.split(','):
        ensure_valid_value('tabstops', vmap, t)
        out.append(vmap[t])
    return out

def insert_hyperlink(para, txt, url):
    # This gets access to the document.xml.rels file and gets a new relation id value
    part = para.part
    r_id = part.relate_to(url, RELATIONSHIP_TYPE.HYPERLINK, is_external=True)

    # Create the w:hyperlink tag and add needed values
    hyperlink = OxmlElement('w:hyperlink')
    hyperlink.set(qn('r:id'), r_id, )

    # Create a w:r element and a new w:rPr element
    new_run = OxmlElement('w:r')
    rPr = OxmlElement('w:rPr')

    # Join all the xml elements together add add the required text to the w:r element
    new_run.append(rPr)
    new_run.text = txt
    hyperlink.append(new_run)

    # Create a new Run object and add the hyperlink into it
    r = para.add_run()
    r._r.append(hyperlink)

    # A workaround for the lack of a hyperlink style (doesn't go purple after using the link)
    # Delete this if using a template that has the hyperlink style in it
    r.font.color.theme_color = MSO_THEME_COLOR_INDEX.HYPERLINK
    #r.font.color.rgb = RGBColor(0x06, 0x45, 0xad)
    r.font.underline = True
    return hyperlink

def assign_numbering(doc, para, anid, start=1, tierlvl=0):
    '''
    how to find abstract_numId (anid), open the .docx and find nav to word/document.xml
    search the relevant keyword in the list that we want to restart (i.e. ListBullet)
    open word/styles.xml and search for the style (ListBullet), look for numId value
    open word/numbering.xml and search the numId="x" value, get abstractNumId value
    '''
    ctn = doc.part.numbering_part.numbering_definitions._numbering
    nxtnid = ctn._next_numId
    num = CT_Num.new(nxtnid, anid)
    num.add_lvlOverride(ilvl=tierlvl).add_startOverride(start)
    ctn._insert_num(num)
    para._p.get_or_add_pPr().get_or_add_numPr().get_or_add_numId().val = nxtnid
    return nxtnid

def set_paranumpr(para, numid=3, ilvl=0):
    npr = para._p.get_or_add_pPr().get_or_add_numPr()
    npr.get_or_add_numId().val = numid
    npr.get_or_add_ilvl().val = ilvl

def make_caption(run, include_heading=0, caption_type='Figure'):
    '''
    make the run a reference caption. set include_heading=1 to do smth like
    Figure x-y, where the x is the heading numbering to follow, and y the n-th figure under heading x
    '''
    r = run._r

    _strseq = f' SEQ {caption_type} \\* ARABIC '
    if isinstance(include_heading, int) and include_heading > 0:
        fldChar = OxmlElement('w:fldChar')
        fldChar.set(qn('w:fldCharType'), 'begin')
        r.append(fldChar)
        instrText = OxmlElement('w:instrText')
        instrText.text = f' STYLEREF {include_heading} \s '
        r.append(instrText)
        fldChar = OxmlElement('w:fldChar')
        fldChar.set(qn('w:fldCharType'), 'end')
        r.append(fldChar)

        nbh = OxmlElement('w:noBreakHyphen')
        r.append(nbh)

        _strseq = f' SEQ Figure \\* ARABIC \s {include_heading}'

    fldChar = OxmlElement('w:fldChar')
    fldChar.set(qn('w:fldCharType'), 'begin')
    r.append(fldChar)
    instrText = OxmlElement('w:instrText')
    instrText.text = _strseq
    r.append(instrText)
    fldChar = OxmlElement('w:fldChar')
    fldChar.set(qn('w:fldCharType'), 'end')
    r.append(fldChar)

def delete_paragraph(para):
    p = para._element
    p.getparent().remove(p)
    p._p = p._element = None

def format_cell(cell, **kwargs):
    if 'color' in kwargs:
        color = parse_color(kwargs.pop('color'))
        sh_elem = parse_xml(r'<w:shd {} w:fill="{}"/>'.format(nsdecls('w'), color))
        cell._tc.get_or_add_tcPr().append(sh_elem)

    if 'align' in kwargs:
        align = kwargs.pop('align')
        for para in cell.paragraphs:
            ikwargs = {'align': align}
            format_paragraph(para, **ikwargs)

    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()

    # check for tag existnace, if none found, then create one
    tcBorders = tcPr.first_child_found_in("w:tcBorders")
    if tcBorders is None:
        tcBorders = OxmlElement('w:tcBorders')
        tcPr.append(tcBorders)

    set_border(tcBorders, **kwargs)

    return cell

def set_figure_dims(figobj, width, height):
    if width is None:
        width = figobj.width
    else:
        width = parse_sizespec(width)

    if height is None:
        # maintain aspect ratio here
        sf = float(width) / float(figobj.width)
        figobj.height = round(figobj.height * sf)
    else:
        # set height
        figobj.height = parse_sizespec(height)

    figobj.width = width

def set_figure_border(figobj, border_width='1.5pt', border_color='000000'):

    strw = str(parse_sizespec(border_width))
    colw = str(parse_color(border_color))
    aln = OxmlElement('a:ln')
    aln.set('w', strw)
    afl = OxmlElement('a:solidFill')
    acl = OxmlElement('a:srgbClr')
    acl.set('val', colw)
    afl.append(acl)
    aln.append(afl)

    inline = figobj._inline
    spPr = inline.graphic.graphicData.pic.spPr
    spPr.append(aln)


    extEff = OxmlElement('wp:effectExtent')
    extEff.set('l', strw)
    extEff.set('t', strw)
    extEff.set('r', strw)
    extEff.set('b', strw)

    ext = inline.extent
    ext.addnext(extEff)


def concat_docx(files):
    if isinstance(files, list) and len(files) > 0:
        main_c = Composer(Document(files[0]))

        for f in files[1:]:
            main_c.append(Document(f))

        return main_c.doc
    raise TypeError('input to concat_docx should be a list')

def set_updatefields(docx, val="true"):
    # https://github.com/elapouya/python-docx-template/issues/151#issuecomment-442722594
    namespace = "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}"
    # add child to doc.settings element
    element_updatefields = lxml.etree.SubElement(
        docx.settings.element, f"{namespace}updateFields"
    )
    element_updatefields.set(f"{namespace}val", val)

def insert_TOC(run, prompt_update):
    _add_lox(run, f'TOC \\o "1-3" \\h \\z \\u', prompt_update)

def insert_LOT(run, prompt_update):
    _add_lox(run, f'TOC \\h \\z \\c "Table"', prompt_update)

def insert_LOF(run, prompt_update=True):
    _add_lox(run, f'TOC \\h \\z \\c "Figure"', prompt_update)

def _add_lox(run, ist, prompt_update):
    # https://github.com/xiaominzhaoparadigm/python-docx-add-list-of-tables-figures/blob/master/LOT.py
    fldChar = OxmlElement('w:fldChar')
    fldChar.set(qn('w:fldCharType'), 'begin')
    # uncomment the following line to prompt user to update the field
    if prompt_update:
        fldChar.set(qn('w:dirty'), 'true')
    instrText = OxmlElement('w:instrText')
    instrText.set(qn('xml:space'), 'preserve')
    instrText.text = ist
    fldChar2 = OxmlElement('w:fldChar')
    fldChar2.set(qn('w:fldCharType'), 'separate')
    fldChar3 = OxmlElement('w:t')
    fldChar3.text = "Right-click to update field."
    fldChar2.append(fldChar3)

    fldChar4 = OxmlElement('w:fldChar')
    fldChar4.set(qn('w:fldCharType'), 'end')

    run._r.append(fldChar)
    run._r.append(instrText)
    run._r.append(fldChar2)
    run._r.append(fldChar4)

def insert_pagenum(section, run, **kwargs):
    _optmap = ('show_total', 'start')
    ensure_valid_attr(_optmap, kwargs.keys())

    fldChar1 = OxmlElement('w:fldChar')
    fldChar1.set(qn('w:fldCharType'), 'begin')

    instrText = OxmlElement('w:instrText')
    instrText.set(qn('xml:space'), 'preserve')
    instrText.text = "PAGE"

    fldChar2 = OxmlElement('w:fldChar')
    fldChar2.set(qn('w:fldCharType'), 'end')

    run._r.append(fldChar1)
    run._r.append(instrText)
    run._r.append(fldChar2)

    if parse_bool(kwargs.get('show_total')):
        # allow 'page x of n' displays
        t2 = OxmlElement('w:t')
        t2.set(qn('xml:space'), 'preserve')
        t2.text = ' of '
        run._r.append(t2)

        fldChar1 = OxmlElement('w:fldChar')
        fldChar1.set(qn('w:fldCharType'), 'begin')

        instrText = OxmlElement('w:instrText')
        instrText.set(qn('xml:space'), 'preserve')
        instrText.text = "NUMPAGES"

        fldChar2 = OxmlElement('w:fldChar')
        fldChar2.set(qn('w:fldCharType'), 'end')

        run._r.append(fldChar1)
        run._r.append(instrText)
        run._r.append(fldChar2)

    pstart = kwargs.get('start')
    if isinstance(pstart, str) and pstart.isnumeric():
        pgNumType =  OxmlElement('w:pgNumType')
        pgNumType.set(qn('w:start'), pstart)

        section._sectPr.append(pgNumType)


def format_table_border(table, **kwargs):
    tblPr = table._tblPr

    # check for tag existence, if none found, then create one
    tblBorders = tblPr.first_child_found_in('w:tblBorders')
    if tblBorders is None:
        tblBorders = OxmlElement('w:tblBorders')
        tblPr.append(tblBorders)

    set_border(tblBorders, **kwargs)
    return table


def indent_table(table, indent):
    # noinspection PyProtectedMember
    tbl_pr = table._element.xpath('w:tblPr')
    if tbl_pr:
        e = OxmlElement('w:tblInd')
        e.set(qn('w:w'), str(indent.twips))
        e.set(qn('w:type'), 'dxa')
        tbl_pr[0].append(e)


def set_border(tbem, **kwargs):
    _edgemap = ('left', 'right', 'top', 'bottom', 'insideH', 'insideV')
    _typemap = ('width', 'line', 'color', 'space', 'shadow')
    _defaults = {
            'width': '12',
            'color': '000000',
            'line': 'single',
            'space': '0',
            'shadow': 'false',
            }

    _optmap = []
    for e in _edgemap:
        for t in _typemap:
            _optmap.append('%s_%s' % (e, t))

    ensure_valid_attr(_optmap, kwargs.keys())
    submap = {
        'width': 'sz',
        'line': 'val',
    }

    # search for edge options, if any edge options exists, all types should be configured
    for e in _edgemap:
        edge_undefined = True
        for opt in kwargs.keys():
            if opt.startswith(e):
                # one of the types on this edge is defined
                edge_undefined = False
                break

        if edge_undefined:
            # skip the rest
            continue

        for t in _typemap:
            _key = '%s_%s' % (e, t)

            if _key in kwargs:
                edat = kwargs.pop(_key)
            else:
                edat = _defaults[t]

            if t == 'width':
                if edat.endswith('pt'):
                    edat = edat[:-2]

            if t == 'color' and edat != 'auto':
                edat = str(parse_color(edat))
                if not edat.startswith('#'):
                    edat = '#%s' % edat

            if t in submap:
                t = submap[t]

            # check tag existence, if does not exist, create one
            tag = 'w:%s' % e
            element = tbem.find(qn(tag))

            if element is None:
                # element does not exist, create one
                element = OxmlElement(tag)
                tbem.append(element)

            element.set(qn('w:%s' % t), edat)

    return tbem

def left_indent_from_level(obj, lvl, indent_p_lvl=Inches(0.4)):
    if isinstance(obj, Paragraph):
        obj.paragraph_format.left_indent = indent_p_lvl * lvl
    elif isinstance(obj, Table):
        indent_table(obj, Twips(indent_p_lvl.twips * lvl))



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import traceback
import io
from mistletoe import block_token

from .. import utils
from .documentx import Renderer as DocxRenderer
from docxcompose.composer import Composer

OPNAME = [
        'mdgen',
        'markdown',
        ]

def can_process(fn):
    return fn.endswith('.md')

def render_file(infile, **kwargs):
    """
    Converts markdown input to the output supported by the given renderer.
    If no renderer is supplied, ``HTMLRenderer`` is used.
    Note that extra token types supported by the given renderer
    are automatically (and temporarily) added to the parsing process.
    """
    with DocxRenderer(**kwargs) as rdr:
        outfile = rdr.render(block_token.Document(infile))
        return outfile

def docx_generate(mdarg, encoding=utils.default_encoding, **kwargs):
    outf = None

    if isinstance(mdarg, list) and len(mdarg) > 0:
        try:
            md = mdarg[0]
            with open(md, 'r', encoding=encoding) as infile:
                main_doc = render_file(infile, **kwargs)
                main_comp = Composer(main_doc)

            for md in mdarg[1:]:
                with open(md, 'r', encoding=encoding) as infile:
                    app_doc = render_file(infile, **kwargs)
                    main_comp.append(app_doc)

            outf = main_comp.doc
        except Exception as e:
            # if something messes up here, we must raise
            # this is a fatal error
            #utils.log.error(f'error on input "{md}" - {e}')
            utils.log.exception(f'error on input "{md}" - {e}')

    else:
        with open(mdarg, 'r', encoding=encoding) as infile:
            outf = render_file(infile, **kwargs)

    #utils.docx_helper.set_updatefields(outf, 'true')
    return outf



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

# renders a docx from markdown
# using mistletoe for markdown parsing and python-docx for docx rendering

import os
import re
import webcolors

from io import StringIO
from itertools import chain
from mistletoe import block_token, span_token
from mistletoe.base_renderer import BaseRenderer

from docx import Document
from .. import utils
from ..utils.docx_helper import (
        format_paragraph, format_run, format_table, format_tabstops,
        format_figure, format_cell, format_section, insert_pagenum, format_table_border,
        insert_hyperlink, assign_numbering, insert_section, insert_hrule, merge_table_cells,
        make_caption, delete_paragraph, insert_LOF, insert_LOT, insert_TOC, left_indent_from_level,
        )

from .format_tag import (
        HorizontalRuleTag, LineBreakTag, PageBreakTag, SectionBreakTag, SectionControlTag,
        BoldTag, ItalicTag, UnderlineTag, StrongTag, EmphasisTag, StrikethroughTag,
        FontTag, ImageTag, CellTag, InsertTabTag, InsertPageNumTag, MergeTag,
        TOCTag, LOTTag, LOFTag,
        )

from .format_tag import (
        CommentBlockTag, AlignBlockTag, TableBlockTag, ParagraphBlockTag,
        FooterBlockTag, HeaderBlockTag, BorderBlockTag, FontBlockTag,
        )

class Renderer(BaseRenderer):
    def __init__(self, docx_template=None, rel_root=None, docx_opts=None, *extras):
        '''
        renders a .docx using docx_template as the template file
        rel_root determine the path to include relative path resources (e.g., images)
        '''
        self._suppress_ptag_stack = [False]
        super().__init__(*chain(([
            HorizontalRuleTag, LineBreakTag, PageBreakTag,
            SectionBreakTag, InsertTabTag, SectionControlTag,
            BoldTag, ItalicTag, UnderlineTag, StrongTag, EmphasisTag, StrikethroughTag,
            FontTag, ImageTag, CellTag, InsertPageNumTag, MergeTag,
            CommentBlockTag, AlignBlockTag, TableBlockTag, ParagraphBlockTag,
            FooterBlockTag, HeaderBlockTag, BorderBlockTag,
            TOCTag, LOTTag, LOFTag, FontBlockTag,
        ]), extras))
        self.rel_root = os.getcwd() if rel_root is None else rel_root
        self.docx_template = docx_template
        self.docx_opts = {}
        if isinstance(docx_opts, dict):
            self.docx_opts = docx_opts

    def render_inner(self, token):
        # inner rendering not supported on DocxRenderer
        # raise NotImplementedError('unsupported render_inner function called')
        for c in token.children:
            self.render(c)

    def render_list(self, token):
        # https://github.com/miyuchina/mistletoe/blob/94022647cd9d80e242db5c93a6567e3155b468bc/mistletoe/block_token.py#L447
        ordered = token.start is not None
        level = self.list_level
        self.list_level += 1

        #if 'List Number' in self.docx.styles and 'List Bullet' in self.docx.styles:
        # old numbering system

        # TODO: fix this abstract_numId map with a more robust implementation
        anid_map = {
            0: 7, # numId 5, absnumId 7
            1: 3, # numId 6, absnumId 3
            2: 2, # numId 7, absnumId 2
        }

        if ordered:
            stpl = 'List Number '
        else:
            stpl = 'List Bullet '

        if level > 0:
            stpl += str(level+1)
        style = stpl.strip()

        for idx, c in enumerate(token.children):
            number = token.start + idx if ordered else None
            # if mistletoe changed the way leader/prepend is computed, this must changed as well

            tos = len(self.paras)
            for ic in c.children:
                ic.docx_style = style
                self.render(ic)
            #self.populate_and_format_paras(c, style=style)

            added = len(self.paras) - tos
            #if added > 0 and ordered:
            if added > 0 and ordered and idx == 0:
                assign_numbering(self.docx, self.paras[tos], anid_map[level], start=number)

        #elif 'List Paragraph' in self.docx.styles:

        #    nid_map = {
        #        True: 3, # ordered use numId = 3
        #        False: 5 # non-ordered use numId = 5, check the .docx document.xml
        #    }

        #    for idx, c in enumerate(token.children):
        #        number = token.start + idx if ordered else 1

        #        tos = len(self.paras)
        #        for ic in c.children:
        #            ic.docx_style = 'List Paragraph'
        #            self.render(ic)
        #        added = len(self.paras) - tos
        #        set_paranumpr(self.paras[tos], nid_map[ordered], level)
        #else:
        #    raise Exception('no valid numbering styles found')

        self.list_level -= 1

    def render_list_item(self, token):
        # handled by list
        #self.render_inner(token)
        raise NotImplementedError("list_item is handled by list rendering")

    def render_link(self, token):
        # TODO: allow alt-txt in links, see mistletoe docs on parser issues (token.title is empty!)
        if len(token.title) < 1:
            ut = token.target
        else:
            ut = token.title
        insert_hyperlink(self.paras[-1], ut, token.target)

    def render_auto_link(self, token):
        insert_hyperlink(self.paras[-1], token.target, token.target)

    def render_image(self, token):
        if token.src.startswith('/'):
            img_path = token.src
        else:
            img_path = os.path.join(self.rel_root, token.src)

        if not os.path.isfile(img_path):
            raise FileNotFoundError(f'image file \'{img_path}\' not found')

        self.runs.append(self.paras[-1].add_run())
        self.inline_shapes.append(self.runs[-1].add_picture(img_path))

        # skip caption if no caption string
        if len(token.title.strip()) < 1:
            return

        self.populate_caption('Figure', token.title)

    def render_block_code(self, token):
        _dxopt = utils.parse_bool(self.docx_opts.get('error_on_blockcode', False))
        if _dxopt:
            raise NotImplementedError('block_codes not supported, use <font name="Lucida Sans Typewriter"></font> instead')
        else:
            par = self.make_paragraph(token)

            rtobj = token.children[0]
            sio = StringIO(rtobj.content)
            for lines in sio:
                self.runs.append(self.paras[-1].add_run())
                self.runs[-1].add_text(lines)
                self.runs[-1].add_break()

    def render_document(self, token):
        # create document from template
        if self.docx_template is None:
            self.docx = Document()
        else:
            self.docx = Document(self.docx_template)
            delete_paragraph(self.docx.paragraphs[-1])

        # paragraph and run stack
        self.list_level = 0 # default root level
        self.heading_level = None # no heading level
        self.runs = []
        self.paras = self.docx.paragraphs.copy()
        self.tables = self.docx.tables.copy()
        self.sections = []
        for s in self.docx.sections:
            self.sections.append(s)
        self.cells = []
        self.inline_shapes = []
        self.render_inner(token)

        # save document
        #self.docx.save(self.out_path)
        return self.docx

    def populate_and_format_runs(self, token, **kwargs):
        # format_tag allows nested formatting <b><i>test</i></b>, italic and bold
        tos = len(self.runs) # top of stack
        self.render_inner(token)
        added = len(self.runs) - tos # new top of stack
        # apply format to all added elements during inner render
        for i in range(added):
            format_run(self.runs[-(i+1)], **kwargs)

    def populate_and_format_paras(self, token, **kwargs):
        tos = len(self.paras)
        self.render_inner(token)
        added = len(self.paras) - tos
        # apply format to all added elements during inner render
        for i in range(added):
            tar = self.paras[-(i+1)]
            format_paragraph(tar, **kwargs)

    def populate_and_format_tables(self, token, **kwargs):
        tos = len(self.tables) # top of stack
        self.render_inner(token)
        added = len(self.tables) - tos # new top of stack
        # add style to all added table within this block
        for i in range(added):
            format_table(self.tables[-(i+1)], **kwargs)

    def render_font_tag(self, token):
        self.populate_and_format_runs(token, **token.format)

    def render_bold_tag(self, token):
        self.populate_and_format_runs(token, bold=True)

    def render_strong(self, token):
        self.populate_and_format_runs(token, style='Strong', bold=True)

    def render_strong_tag(self, token):
        self.populate_and_format_runs(token, style='Strong')

    def render_italic_tag(self, token):
        self.populate_and_format_runs(token, italic=True)

    def render_emphasis(self, token):
        self.populate_and_format_runs(token, style='Emphasis', italic=True)

    def render_emphasis_tag(self, token):
        self.populate_and_format_runs(token, style='Emphasis')

    def render_underline_tag(self, token):
        self.populate_and_format_runs(token, underline=True)

    def render_strikethrough_tag(self, token):
        self.populate_and_format_runs(token, strike=True)

    def render_strikethrough(self, token):
        self.populate_and_format_runs(token, strike=True)

    def render_heading(self, token):
        # assume that heading has no additional child
        self.heading_level = token.level
        tar = self.docx.add_heading(level=token.level).clear()
        self.paras.append(tar)
        self.render_inner(token)
        self.auto_left_indent(tar, token.level-1)

    def make_paragraph(self, token):
        render_ptr = self.docx
        if hasattr(token, 'render_to') and isinstance(token.render_to, str):
            if token.render_to == 'header':
                render_ptr = self.sections[-1].header

            elif token.render_to == 'footer':
                render_ptr = self.sections[-1].footer
        else:
            token.render_to = 'body'

        if hasattr(token, 'docx_style') and isinstance(token.docx_style, str):
            par = render_ptr.add_paragraph(style=token.docx_style)
        else:
            par = render_ptr.add_paragraph()
        par.clear()

        self.paras.append(par)
        return par

    def render_paragraph(self, token):
        # for all span tokens
        par = self.make_paragraph(token)
        self.render_inner(token)
        self.auto_left_indent(par)

    def auto_left_indent(self, obj, lvl = None):
        _dxopt = utils.parse_docx_sizespec(self.docx_opts.get('auto_left_indent', None))
        if not _dxopt:
            return
        if not isinstance(self.heading_level, int):
            # no headings yet, left most
            lvl = 0
        elif not isinstance(lvl, int):
            # there are headings, but lvl is not an int (default)
            lvl = self.heading_level
        left_indent_from_level(obj, lvl, _dxopt)

    def render_horizontal_rule_tag(self, token):
        insert_hrule(self.paras[-1], token.format_value)

    def render_thematic_break(self, token):
        insert_hrule(self.paras[-1])

    def render_page_break_tag(self, token):
        self.docx.add_page_break()

    def render_toc_tag(self, token):
        self.runs.append(self.paras[-1].add_run())
        _dxopt = utils.parse_bool(self.docx_opts.get('prompt_updatefield', True))
        insert_TOC(self.runs[-1], _dxopt)

    def render_lot_tag(self, token):
        self.runs.append(self.paras[-1].add_run())
        _dxopt = utils.parse_bool(self.docx_opts.get('prompt_updatefield', True))
        insert_LOT(self.runs[-1], _dxopt)

    def render_lof_tag(self, token):
        self.runs.append(self.paras[-1].add_run())
        _dxopt = utils.parse_bool(self.docx_opts.get('prompt_updatefield', True))
        insert_LOF(self.runs[-1], _dxopt)

    def render_raw_text(self, token):
        # add run to last added paragraph
        self.runs.append(self.paras[-1].add_run())
        self.runs[-1].add_text(token.content)

    def render_line_break(self, token):
        self.runs.append(self.paras[-1].add_run())
        self.runs[-1].add_break()

    def render_line_break_tag(self, token):
        self.runs.append(self.paras[-1].add_run())
        self.runs[-1].add_break()

    def render_insert_tab_tag(self, token):
        # insert a tab character
        if len(self.runs) < 1:
            self.runs.append(self.paras[-1].add_run())
        self.runs[-1].add_tab()

    def render_insert_page_num_tag(self, token):
        self.runs.append(self.paras[-1].add_run())
        insert_pagenum(self.sections[-1], self.runs[-1], **token.format)

    def render_section_break_tag(self, token):
        # insert new section
        new_section = insert_section(self.docx, token.format_value)
        new_section.header.is_linked_to_previous = False
        new_section.footer.is_linked_to_previous = False
        self.sections.append(new_section)

    def render_section_control_tag(self, token):
        # make adjustments to current section
        format_section(self.sections[-1], **token.format)

    def render_comment_block_tag(self, token):
        # do nothing
        pass

    def render_table(self, token):
        if len(token.children) < 1:
            return
        ncol = len(token.children[0].children)
        tbltar = self.docx.add_table(0, ncol)
        self.tables.append(tbltar)

        if hasattr(token, 'header') and token.header is not None:
            token.header.is_header = True
            self.render(token.header)

        for row in token.children:
            row.is_header = False
            self.render(row)

        self.auto_left_indent(tbltar)


    def render_table_row(self, token):
        row = self.tables[-1].add_row()
        for cidx, col in enumerate(token.children):
            if cidx >= len(row.cells):
                utils.log.warn('table has %d columns but a row has %d columns: %s' % (len(row.cells), cidx+1, str([c.content for c in col.children])))
                continue
            self.cells.append(row.cells[cidx])
            self.render(col)

    def render_table_cell(self, token):
        self.paras.append(self.cells[-1].paragraphs[0])  # add default para on cell
        self.render_inner(token)

    def render_table_block_tag(self, token):
        # additional valid opts
        if isinstance(token.format.get('caption'), str):
            kwargs = {
                'align': token.format.pop('caption_align', None)
            }
            self.populate_caption('Table', token.format.get('caption'), **kwargs)
            token.format.pop('caption')
        self.populate_and_format_tables(token, **token.format)

    def render_cell_tag(self, token):
        if len(self.cells) < 1:
            return

        self.render_inner(token)
        format_cell(self.cells[-1], **token.format)

    def render_merge_tag(self, token):
        if len(self.tables) < 1:
            return

        merge_table_cells(self.tables[-1], **token.format)

    def render_border_block_tag(self, token):
        tos = len(self.tables)
        self.render_inner(token)
        added = len(self.tables) - tos
        for i in range(added):
            target = self.tables[-(i+1)]
            format_table_border(target, **token.format)

    def render_align_block_tag(self, token):
        self.populate_and_format_paras(token, align=token.format_value)

    def populate_caption(self, caption_type, caption_string, **kwargs):
        tcpar = self.docx.add_paragraph(style='Caption').clear()
        self.paras.append(tcpar)
        # self.auto_left_indent(tcpar)  # uncomment this to auto-indent captions
        format_paragraph(tcpar, **kwargs)
        _dxopt = int(self.docx_opts.get('caption_prefix_heading', 0))
        make_caption(tcpar.add_run(f'{caption_type} '), _dxopt, caption_type)

        tcrun = tcpar.add_run(f': {caption_string}')

    def render_image_tag(self, token):
        tos = len(self.inline_shapes)
        self.render_inner(token)
        added = len(self.inline_shapes) - tos
        for i in range(added):
            target = self.inline_shapes[-(i+1)]
            format_figure(target, **token.format)

    def render_paragraph_block_tag(self, token):
        self.populate_and_format_paras(token, **token.format)

    def render_header_block_tag(self, token):
        _tsalign = format_tabstops(token.format.get('tabstops'), self.sections[-1])
        utils.set_attr_recursively(token, block_token.Paragraph, 'render_to', 'header')
        self.populate_and_format_paras(token, tabstops=_tsalign)

    def render_footer_block_tag(self, token):
        _tsalign = format_tabstops(token.format.get('tabstops'), self.sections[-1])
        utils.set_attr_recursively(token, block_token.Paragraph, 'render_to', 'footer')
        self.populate_and_format_paras(token, tabstops=_tsalign)

    def render_font_block_tag(self, token):
        self.populate_and_format_runs(token, **token.format)



from setuptools import find_packages, setup

from text_office import version
_packages = find_packages()

setup(
    name='text-office',
    version=version,
    description='a python module to create office documents from text-based files (e.g., markdown, xml)',
    packages=_packages,
    author='Chia Jason',
    author_email='chia_jason96@live.com',
    url='https://github.com/toranova/text-office/',
    download_url='https://github.com/ToraNova/text-office/archive/refs/tags/v%s.tar.gz' % version,
    license='AGPL-3.0-or-later',
    include_package_data=True,
    data_files=[('boiler_templates', ['boiler_templates/dtpt-1.md'])],
    zip_safe=False,
    keywords = ['docx', 'md', 'xml'],
    install_requires=[
        'cvss==2.5',
        'docxcompose==1.3.4',
        'lxml==4.9.3',
        'mistletoe==1.1.0',
        'natsort==8.1.0',
        'python-docx==0.8.11',
        'six==1.16.0',
        'webcolors==1.11.1',
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
    scripts=["text-office.py", "boilergen.py"],
)



#!/usr/bin/env python
'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import io
import argparse
import datetime
from os import listdir
from natsort import natsorted
from os.path import isfile, isdir, join
from docx import Document
from docx.document import Document as _DOCX
from text_office import (
        version,
        md,
        utils,
        )

JOIN_OP = ['djoin', 'dconcat']
LIST_OP = ['listonly', 'nogen']
LSTY_OP = ['show_default_styles']
MKTP_OP = ['mktpl']
MERG_OP = ['tmerge']

parser = argparse.ArgumentParser()
# single
parser.add_argument('inputs', help='input files', type=str, nargs='*')
parser.add_argument('-f', '--manifest', help='processing manifest', type=str)
parser.add_argument('--ascending', help='sort in ascending order (if sorting)', action='store_true')
parser.add_argument('--nosort', help='do not sort input naturally', action='store_true')
parser.add_argument('-op', '--operation', help=f'type of operation to do: ({md.OPNAME}), {JOIN_OP}, {LIST_OP}, {MKTP_OP}, {LSTY_OP}, {MERG_OP}', type=str, default='generate')
parser.add_argument('-o', '--output', help='output docx path', type=str, default='output.docx')
parser.add_argument('-t', '--template', help='template docx path', type=str)
parser.add_argument('-dxopt', '--docx_opts', help='key-value pair of docx options (e.g., caption_prefix_heading=1, prompt_updatefield=no)', type=str)
parser.add_argument('--rel_root', help='relative root (for images, attachments)', type=str)
parser.add_argument('--version', help='show version number', action='store_true')
args = parser.parse_args()

if args.version:
    # show version then exit
    print('text-office', version)
    exit(0)

docx = None

# TODO: allow diff modules (e.g., substitute)
_module = md

inlist = []

if args.manifest is not None:
    # using a manifest file
    if not isfile(args.manifest):
        utils.log.critical('cannot open manifest for processing')
        exit(1)

    with open(args.manifest, 'r') as mfile:
        for line in mfile:
            line = line.strip()
            if line.startswith('#'):
                # ignore comments
                continue

            if line.isspace() or len(line) < 1:
                # ignore whitelines
                continue

            if not isfile(line):
                # raise error
                utils.log.error(f'cannot open specified file in manifest "{args.manifest}" for processing: {line}')
                continue

            if _module.can_process(line):
                inlist.append(line)
else:
    # specifying directly from args
    for inp in args.inputs:
        if isfile(inp):
            # is a file
            inlist.append(inp)
        else:
            for _file in listdir(inp):
                _filepath = join(inp, _file)
                if isfile(_filepath) and _module.can_process(_filepath):
                    inlist.append(_filepath)

if args.nosort or args.operation in MERG_OP:
    # don't sort
    pass
else:
    # sort
    inlist = natsorted(inlist, reverse=not args.ascending)


if args.operation in LIST_OP:
    # list target files to process without processing
    utils.log.info("the following markdown will be processed in order:")
    for i in inlist:
        utils.log.info(i)

elif args.operation in JOIN_OP:
    # concatenate docx
    docx = utils.docx_helper.concat_docx(inlist)

elif args.operation in MKTP_OP:
    # create template docx
    docx = Document()

elif args.operation in LSTY_OP:

    _docx = Document()
    for s in _docx.styles:
        print(s)
    docx = None

elif args.operation in MERG_OP:
    if args.output.endswith('.docx'):
        args.output = args.output[:-5] + '.md'

    with open(args.output, 'w', encoding=utils.default_encoding) as outfile:
        for md in inlist:
            with open(md, 'r', encoding=utils.default_encoding) as infile:
                outfile.write(infile.read())

else:
    if len(inlist) < 1:
        utils.log.critical('no inputs')
        exit(1)

    docx_opts = {}
    if args.docx_opts is not None:
        docx_opts = utils.parse_kv_pairs(args.docx_opts)

    # generate
    docx = _module.docx_generate(
            inlist,
            docx_template=args.template,
            rel_root=args.rel_root,
            docx_opts=docx_opts,
            )

if isinstance(docx, _DOCX):
    utils.log.info(f'docx generated at {args.output}')
    docx.save(args.output)



'''
Copyright (C) 2023 ToraNova

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import argparse
import os
from string import Template
from .. import utils

def generate(inargs):
    parser = argparse.ArgumentParser(description="vapt boilerplate module")
    parser.add_argument('template', help='template (relative to boiler_templates dir. in project root)', type=str)
    parser.add_argument('name', help='name of finding', type=str)
    parser.add_argument('-c', '--cvss', help="cvss vector of finding", type=str)
    args = parser.parse_args(inargs)

    severity, score, score_print, vector, color = utils.vapt_helper.get_ssvc(args.cvss)

    finding_name = args.name
    clean_name = args.name.casefold().translate(dict.fromkeys(map(ord, u' \n#/\\()[]{}<>-.')))
    strscore = str(score).replace('.','')
    mdfile_name = f'{strscore}_{clean_name}.md'

    tplfn = utils.ensure_template_file(args.template, auto_add_suffix='.md')
    utils.log.info(f'creating boilerplate markdown with: {tplfn} {severity} {score} {vector} {color}')

    with open(tplfn, 'r') as tpl:
        tplstr = Template(tpl.read())

    res = tplstr.substitute(
        finding_name=finding_name,
        severity=severity,
        score_print=score_print,
        vector=vector,
        color=color
    )

    with open(mdfile_name, 'a+') as out:
        out.write(res)
